/////////////////////////////////////////////////////////////////////////////
//                                                                         //
//                INTEL CORPORATION PROPRIETARY INFORMATION                //
//   This software is supplied under the terms of a license agreement or   //
//  nondisclosure agreement with Intel Corporation and may not be copied   //
//   or disclosed except in accordance with the terms of that agreement.   //
//       Copyright (c) 2003 Intel Corporation. All Rights Reserved.        //
//                                                                         //
//  File:      pnlExInferenceEngine.hpp                                    //
//                                                                         //
//  Purpose:   CExInferenceEngine class definition                         //
//                                                                         //
//  Author(s):                                                             //
//                                                                         //
/////////////////////////////////////////////////////////////////////////////

#ifndef __PNLEXINFERENCEENGINE_HPP__
#define __PNLEXINFERENCEENGINE_HPP__

#include "pnlInferenceEngine.hpp"
#include "pnlNaiveInferenceEngine.hpp"
#include "pnlTabularPotential.hpp"
#include "pnlGaussianPotential.hpp"

PNL_BEGIN

enum EExInfEngineFlavour
{
    PNL_EXINFENGINEFLAVOUR_DISCONNECTED = 1 << 0,
    PNL_EXINFENGINEFLAVOUR_UNSORTED = 1 << 1,
    PNL_EXINFENGINEFLAVOUR_JTREEKLUDGE = 1 << 2,

    PNL_EXINFENGINEFLAVOUR_DISCONNECTED_UNSORTED = PNL_EXINFENGINEFLAVOUR_DISCONNECTED | PNL_EXINFENGINEFLAVOUR_UNSORTED
};

#define PNL_IS_EXINFENGINEFLAVOUR_DISCONNECTED( flav ) ((flav) & PNL_EXINFENGINEFLAVOUR_DISCONNECTED)
#define PNL_IS_EXINFENGINEFLAVOUR_UNSORTED( flav ) ((flav) & PNL_EXINFENGINEFLAVOUR_UNSORTED)
#define PNL_IS_EXINFENGINEFLAVOUR_JTREEKLUDGE( flav ) ((flav) & PNL_EXINFENGINEFLAVOUR_JTREEKLUDGE)

template< class INF_ENGINE, class MODEL = CBNet,
          EExInfEngineFlavour FLAV = PNL_EXINFENGINEFLAVOUR_DISCONNECTED_UNSORTED,
          class FALLBACK_ENGINE1 = CNaiveInfEngine, class FALLBACK_ENGINE2 = INF_ENGINE >
class PNL_API CExInfEngine: public CInfEngine
{
public:
    static CExInfEngine *Create( CStaticGraphicalModel const *model );

    inline CStaticGraphicalModel const *GetModel() const;

    inline static CNodeType const *GetObsGauNodeType();

    inline static CNodeType const *GetObsTabNodeType();

    void EnterEvidence( CEvidence const *evidence,
                        int maximize = 0,
                        int sumOnMixtureNode = 1 );

#ifdef PNL_OBSOLETE
    void MarginalNodes( int const *query, int querySize,
                        int notExpandJPD = 0 );
#endif

    void MarginalNodes( intVector const &queryNdsIn,
                        int notExpandJPD = 0 );

    CPotential const *GetQueryJPD() const;

    CEvidence const *GetMPE() const;

    ~CExInfEngine();

protected:
    CExInfEngine( CStaticGraphicalModel const *pGM );

    void PurgeEvidences();

    EInfTypes inf_type;
    CStaticGraphicalModel const *graphical_model;

    pnlVector< CInfEngine * > engines;
    pnlVector< MODEL * > models;
    pnlVector< CGraph * > graphs;
    pnlVector< CEvidence * > evs;
    intVector orig2comp, orig2idx;
    intVecVector decomposition;
    intVector active_components;
    intVector saved_query;

    CEvidence const *evidence;
    bool evidence_mine;

    pnlVector< pnlVector< CNodeType > > node_types;
    intVecVector node_assoc;

    int maximize;
    mutable CPotential *query_JPD;
    mutable CEvidence *MPE_ev;
};

#ifndef SWIG

template< class INF_ENGINE, class MODEL, EExInfEngineFlavour FLAV, class FALLBACK_ENGINE1, class FALLBACK_ENGINE2 >
inline CExInfEngine< INF_ENGINE, MODEL, FLAV, FALLBACK_ENGINE1, FALLBACK_ENGINE2 > *CExInfEngine< INF_ENGINE, MODEL, FLAV, FALLBACK_ENGINE1, FALLBACK_ENGINE2 >::Create( CStaticGraphicalModel const *gm )
{
    return new( CExInfEngine< INF_ENGINE, MODEL, FLAV, FALLBACK_ENGINE1, FALLBACK_ENGINE2 > )( gm );
}

template< class INF_ENGINE, class MODEL, EExInfEngineFlavour FLAV, class FALLBACK_ENGINE1, class FALLBACK_ENGINE2 >
inline CStaticGraphicalModel const *CExInfEngine< INF_ENGINE, MODEL, FLAV, FALLBACK_ENGINE1, FALLBACK_ENGINE2 >::GetModel() const
{
    return graphical_model;
}

template< class INF_ENGINE, class MODEL, EExInfEngineFlavour FLAV, class FALLBACK_ENGINE1, class FALLBACK_ENGINE2 >
inline CNodeType const *CExInfEngine< INF_ENGINE, MODEL, FLAV, FALLBACK_ENGINE1, FALLBACK_ENGINE2 >::GetObsTabNodeType()
{
    return INF_ENGINE::GetObsTabNodeType();
}

template< class INF_ENGINE, class MODEL, EExInfEngineFlavour FLAV, class FALLBACK_ENGINE1, class FALLBACK_ENGINE2 >
inline CNodeType const *CExInfEngine< INF_ENGINE, MODEL, FLAV, FALLBACK_ENGINE1, FALLBACK_ENGINE2 >::GetObsGauNodeType()
{
    return INF_ENGINE::GetObsGauNodeType();
}

template< class INF_ENGINE, class MODEL, EExInfEngineFlavour FLAV, class FALLBACK_ENGINE1, class FALLBACK_ENGINE2 >
void CExInfEngine< INF_ENGINE, MODEL, FLAV, FALLBACK_ENGINE1, FALLBACK_ENGINE2 >::PurgeEvidences()
{
    int i;

    if ( evidence_mine )
    {
        delete( evidence );
        evidence = 0;
        evidence_mine = false;
    }
    for ( i = evs.size(); i--; )
    {
        delete( evs[i] );
        evs[i] = 0;
    }
}

template< class INF_ENGINE, class MODEL, EExInfEngineFlavour FLAV, class FALLBACK_ENGINE1, class FALLBACK_ENGINE2 >
CExInfEngine< INF_ENGINE, MODEL, FLAV, FALLBACK_ENGINE1, FALLBACK_ENGINE2 >::CExInfEngine( CStaticGraphicalModel const *gm )
    : CInfEngine( itEx, gm ), evidence_mine( false ),
      maximize( 0 ), MPE_ev( 0 ), query_JPD( 0 ), graphical_model( gm )
{
    int i, j, k;
    intVector dom;
    intVector conv;
    CFactor *fac;

    PNL_MAKE_LOCAL( CGraph *, gr, gm, GetGraph() );
    PNL_MAKE_LOCAL( int, sz, gr, GetNumberOfNodes() );

    gr->GetConnectivityComponents( &decomposition );

    for ( i = decomposition.size(); i--; )
    {
        std::sort( decomposition[i].begin(), decomposition[i].end() );
    }
    if ( PNL_IS_EXINFENGINEFLAVOUR_UNSORTED( FLAV ) )
    {
        gr->GetTopologicalOrder( &conv );
    }

    orig2comp.resize( sz );
    orig2idx.resize( sz );

    for ( k = 2; k--; )
    {
        for ( i = decomposition.size(); i--; )
        {
            for ( j = decomposition[i].size(); j--; )
            {
                orig2comp[decomposition[i][j]] = i;
                orig2idx[decomposition[i][j]] = j;
            }
        }

        if ( PNL_IS_EXINFENGINEFLAVOUR_UNSORTED( FLAV ) && k )
        {
            for ( i = sz; i--; )
            {
                decomposition[orig2comp[conv[i]]][orig2idx[conv[i]]] = i;
            }
        }
        else
        {
            break;
        }
    }

    graphs.resize( decomposition.size() );
    models.resize( decomposition.size() );
    engines.resize( decomposition.size() );

    for ( i = decomposition.size(); i--; )
    {
        graphs[i] = gr->ExtractSubgraph( decomposition[i] );
#if 0
        std::cout << "graph " << i << std::endl;
        graphs[i]->Dump();
#endif
    }
    node_types.resize( decomposition.size() );
    node_assoc.resize( decomposition.size() );
    for ( i = 0, k = 0; i < decomposition.size(); ++i )
    {
        node_types[i].resize( decomposition[i].size() );
        node_assoc[i].resize( decomposition[i].size() );
        for ( j = 0; j < decomposition[i].size(); ++j )
        {
            node_types[i][j] = *gm->GetNodeType( decomposition[i][j] );
            node_assoc[i][j] = j;
        }
    }
    for ( i = decomposition.size(); i--; )
    {
        models[i] = MODEL::Create( decomposition[i].size(), node_types[i], node_assoc[i], graphs[i] );
    }
    for ( i = 0; i < gm->GetNumberOfFactors(); ++i )
    {
        fac = gm->GetFactor( i );
        fac->GetDomain( &dom );
#if 0
        std::cout << "Ex received orig factor" << std::endl;
        fac->GetDistribFun()->Dump();
#endif
        k = orig2comp[dom[0]];
        for ( j = dom.size(); j--; )
        {
            dom[j] = orig2idx[dom[j]];
        }
        fac = CFactor::CopyWithNewDomain( fac, dom, models[k]->GetModelDomain() );
#if 0
        std::cout << "Ex mangled it to" << std::endl;
        fac->GetDistribFun()->Dump();
#endif
        models[k]->AttachFactor( fac );
    }
    for ( i = decomposition.size(); i--; )
    {
        switch ( decomposition[i].size() )
        {
        case 1:
            engines[i] = FALLBACK_ENGINE1::Create( models[i] );
            continue;
        case 2:
            engines[i] = FALLBACK_ENGINE2::Create( models[i] );
            continue;
        default:
            engines[i] = INF_ENGINE::Create( models[i] );
        }
    }
}

// current implementation works inefficiently when evidence resides
// only on small fraction of connectivity components of the model
template< class INF_ENGINE, class MODEL, EExInfEngineFlavour FLAV, class FALLBACK_ENGINE1, class FALLBACK_ENGINE2 >
void CExInfEngine< INF_ENGINE, MODEL, FLAV, FALLBACK_ENGINE1, FALLBACK_ENGINE2 >::EnterEvidence( CEvidence const *evidence,
                                                                                                 int maximize, int sumOnMixtureNode )
{
    int i, j, k, m;
    intVector nodes;
    pConstValueVector values;
    pConstNodeTypeVector node_types;
    intVecVector dispenser;
    valueVecVector dispenser_values;

    PurgeEvidences();
    this->evidence = evidence;
    this->maximize = maximize;

    evidence->GetObsNodesWithValues( &nodes, &values, &node_types );
    dispenser.resize( decomposition.size() );
    dispenser_values.resize( decomposition.size() );
    for ( i = 0, m = 0; i < nodes.size(); ++i )
    {
        dispenser[orig2comp[nodes[i]]].push_back( orig2idx[nodes[i]] );
        k = node_types[i]->IsDiscrete() ? 1 : node_types[i]->GetNodeSize();
        for ( j = 0; j < k; ++j )
        {
            dispenser_values[orig2comp[nodes[i]]].push_back( *values[m++] );
        }
    }
    evs.resize( decomposition.size() );
    for ( i = decomposition.size(); i--; )
    {
        evs[i] = 0;
        if ( dispenser[i].size() )
        {
            evs[i] = CEvidence::Create( models[i], dispenser[i], dispenser_values[i] );
#if 0
            engines[i]->GetModel()->GetGraph()->Dump();
#endif
            engines[i]->EnterEvidence( evs[i], maximize );
        }
    }
}

template< class INF_ENGINE, class MODEL, EExInfEngineFlavour FLAV, class FALLBACK_ENGINE1, class FALLBACK_ENGINE2 >
CExInfEngine< INF_ENGINE, MODEL, FLAV, FALLBACK_ENGINE1, FALLBACK_ENGINE2 >::~CExInfEngine()
{
    int i;

    delete( MPE_ev );
    delete( query_JPD );

    PurgeEvidences();
    for ( i = engines.size(); i--; )
    {
        delete( engines[i] );
    }
    for ( i = models.size(); i--; )
    {
        delete( models[i] );
    }
}

// current implementation works not very inefficiently when evidence resides
// only on small fraction of connectivity components of the model
template< class INF_ENGINE, class MODEL, EExInfEngineFlavour FLAV, class FALLBACK_ENGINE1, class FALLBACK_ENGINE2 >
void CExInfEngine< INF_ENGINE, MODEL, FLAV, FALLBACK_ENGINE1, FALLBACK_ENGINE2 >::MarginalNodes( int const *query, int querySize, int notExpandJPD )
{
    int i, j;
    intVecVector dispenser;

    if( querySize == 0 )
    {
        PNL_THROW( CInconsistentSize, "query nodes vector should not be empty" );
    }

    saved_query.assign( query, query + querySize );

    if ( evidence == 0 )
    {
        evidence_mine = true;
        EnterEvidence( evidence = CEvidence::Create( graphical_model, intVector(), valueVector() ) );
    }

    active_components.resize( 0 );
    dispenser.resize( decomposition.size() );

    for ( i = 0; i < querySize; ++i )
    {
        dispenser[orig2comp[query[i]]].push_back( orig2idx[query[i]] );
    }

    for ( i = decomposition.size(); i--; )
    {
        if ( dispenser[i].size() )
        {
            active_components.push_back( i );
            if ( evs[i] == 0 )
            {
                evs[i] = CEvidence::Create( models[i], intVector(), valueVector() );
                engines[i]->EnterEvidence( evs[i], maximize );
            }
            engines[i]->MarginalNodes( &dispenser[i].front(), dispenser[i].size(), notExpandJPD );
        }
    }
}

// current implementation works not very inefficiently when evidence resides
// only on small fraction of connectivity components of the model
template< class INF_ENGINE, class MODEL, EExInfEngineFlavour FLAV, class FALLBACK_ENGINE1, class FALLBACK_ENGINE2 >
void CExInfEngine< INF_ENGINE, MODEL, FLAV, FALLBACK_ENGINE1, FALLBACK_ENGINE2 >::MarginalNodes( intVector const &queryNds, int notExpandJPD )
{
    MarginalNodes( &queryNds.front(), queryNds.size(), notExpandJPD );
}

// current implementation works not very inefficiently when evidence resides
// only on small fraction of connectivity components of the model
template< class INF_ENGINE, class MODEL, EExInfEngineFlavour FLAV, class FALLBACK_ENGINE1, class FALLBACK_ENGINE2 >
CEvidence const *CExInfEngine< INF_ENGINE, MODEL, FLAV, FALLBACK_ENGINE1, FALLBACK_ENGINE2 >::GetMPE() const
{
    int i, j;
    intVector conv_nodes;
    valueVector conv_values;

    intVector nodes;
    pConstValueVector values;
    pConstNodeTypeVector node_types;

    CEvidence const *ev;

    for ( i = active_components.size(); i--; )
    {
        ev = engines[active_components[i]]->GetMPE();
        ev->GetObsNodesWithValues( &nodes, &values, &node_types );
        for ( j = nodes.size(); j--; )
        {
            conv_nodes.push_back( decomposition[active_components[i]][nodes[j]] );
            conv_values.push_back( Value( *values[j] ) );
        }
    }

    return MPE_ev = CEvidence::Create( graphical_model, conv_nodes, conv_values );
}

template< class INF_ENGINE, class MODEL, EExInfEngineFlavour FLAV, class FALLBACK_ENGINE1, class FALLBACK_ENGINE2 >
CPotential const *CExInfEngine< INF_ENGINE, MODEL, FLAV, FALLBACK_ENGINE1, FALLBACK_ENGINE2 >::GetQueryJPD() const
{
    int i, j;
    CPotential *pot1;
    CPotential const *pot2;
    intVector dom;

    if ( query_JPD == 0 )
    {
        if ( active_components.size() > 1 )
        {
            if ( graphical_model->GetNodeType( saved_query[0] )->IsDiscrete() )
            {
                query_JPD = CTabularPotential::CreateUnitFunctionDistribution( saved_query, graphical_model->GetModelDomain() );
            }
            else
            {
                query_JPD = CGaussianPotential::CreateUnitFunctionDistribution( saved_query, graphical_model->GetModelDomain() );
            }

            for ( i = active_components.size(); i--; )
            {
                pot2 = engines[active_components[i]]->GetQueryJPD();
                pot2->GetDomain( &dom );
                for ( j = dom.size(); j--; )
                {
                    dom[j] = decomposition[active_components[i]][dom[j]];
                }
                pot1 = (CPotential *)CPotential::CopyWithNewDomain( pot2, dom, graphical_model->GetModelDomain() );
                *query_JPD *= *pot1;
                delete( pot1 );
            }

#if 0
            query_JPD->Normalize();
#endif
        }
        else
        {
            pot2 = engines[active_components[0]]->GetQueryJPD();
            pot2->GetDomain( &dom );
            for ( j = dom.size(); j--; )
            {
                dom[j] = decomposition[active_components[0]][dom[j]];
            }
            query_JPD = (CPotential *)CPotential::CopyWithNewDomain( pot2, dom, graphical_model->GetModelDomain() );
        }
    }

    return query_JPD;
}

#endif

PNL_END

#endif // __PNLINFERENCEENGINE_HPP__
