#include "ETIGreedy.h"

int main(int argc, char **argv)
{
  // make sure parameters are specified to the program
  if (argc == 1) {
    cerr << "usage: getipp dataset support epsCol epsRow" << endl;
    exit(1);
  }    

  DataSet data(argv[1]);	
  ETIGreedy miner(&data, atof(argv[3]), atof(argv[4]), atoi(argv[2]));
  miner.MineETIs();

  return 0;
}

ETIGreedy::ETIGreedy(DataSet *data, double epsc, double epsr, int minsup):
  dataSet(data)
{
  if (epsc >= 0 && epsc <= 1.0)
    epsCol = epsc;
  else {
    epsCol = 0;
    cerr << "epsCol should be in the range [0..1]" << endl;
  }

  if (epsr >= 0 && epsr <= 1.0)
    epsRow = epsr;		//
  else {
    epsRow = 0;
    cerr << "epsRow should be in the range [0..1]" << endl;
  }

  minSup = minsup;
  epsPP = epsr;
}

void ETIGreedy::MineETIs()
{
  bool foundSome = FindSomeETIs(1);
  while (foundSome) {
    foundSome = FindSomeETIs(2);
  }

  // do some post-processing
  dataSet->SetAllRowsAvailable();
  vector<Bitmap> verifiedETIs;
  int size = etis.size();
  int *zeros = new int[dataSet->GetTransactionCount()];
  for (int i = 0; i < size; i++) {
    double sigma = ComputeSigma(&(etis[i]), zeros, 1, false);

		// verify the strong ETI condition
		if (sigma <= epsCol) {
			if (HasStrongSupport(etis[i])) {
      	verifiedETIs.push_back(etis[i]);
	      PrintItems(&(etis[i]));
   	 	}
		}
  }
  etis.clear();
}

// the greedy growing strategy
bool ETIGreedy::FindSomeETIs(int lambda)
{
  vector<Bitmap *> candidates;
  int transCount = dataSet->GetTransactionCount();
  int itemCount = dataSet->GetItemCount();
  int *zeros = new int[transCount];
  bool anyAdded = false;

  ComputePruningSupport();

  // find the first item that meets the pruning threshold
  // this assumes that the data is sorted in increasing support
  // test for the bad things that happen if no items meet the criterion
  int firstItem = itemCount;
  for (int item = 0; item < itemCount; item++) {
    if (dataSet->GetItemBitmap(item)->CountOnes() >= (pruningSupport[1] / lambda)) {
      firstItem = item;
      break;
    }
  } 
  if (firstItem == itemCount) {
    cout << "No items were frequent enough to make any ETIs" << endl;
    return false;
  }
  //firstItem = 0;

  // for each of the frequent enough singleton seeds, just keep on grabbing
  // the best item to add, and create some ETIs in that manner
  // this is probably not efficient, but it sure is the simplest way to think
  // of things
  for (int seed = firstItem; seed < itemCount; seed++) {
    //cout << "-------------------------------------" << endl;
    // make an itemset
    Bitmap items = Bitmap(itemCount);
    items.Clear();
    items.SetBit(seed, true);

    // keep add items as long as possible
    bool keepTrying = true;
    double bestSigma = 1.01;
    while (keepTrying) {
      bestSigma = 1.01;
      int bestIndex = -1;
      bool possibleExtensions = false;
      //  cout << "-------------------------------------" << endl;

      // find the best extension by adding any possible item
      for (int extend = firstItem; extend < itemCount; extend++) {
	if (!items.GetBit(extend)) {
	  ComputeZeros(&items, zeros);
	  //double coverage = ComputeCoverage(&items, extend, zeros);

	  // figure out if this is still a weak ETI
	  items.SetBit(extend, true);
	  double sigma = ComputeSigma(&items, zeros, lambda, false);
	  bool definesETI = (sigma <= epsCol);

	  // cout << "extend: " << dataSet->GetActualIndex(extend)
	  //     << ", sigma: " << sigma << endl;

	  if (sigma < bestSigma && definesETI) {	    
	    //	    bestCoverage = coverage;
	    bestSigma = sigma;
	    bestIndex = extend;
	    possibleExtensions = true;
	  }
	  items.SetBit(extend, false);
	}
      }
         
      if (possibleExtensions) {	
	//cout << "made choice: " << dataSet->GetActualIndex(bestIndex) << endl;
	//cout << bestSigma << endl;
	items.SetBit(bestIndex, true);
      } else {	
	keepTrying = false;
      }
    }
   
    bool added = AddETIToList(items);
    if (added) {	 
      anyAdded = true;
      //PrintItems(&items);
    }
    
  }

  // determine which rows are covered by at least one ETI
  // and remove those from consideration
  int rows = dataSet->GetTransactionCount();
  int etiCount = etis.size();
  for (int r = 0; r < rows; r++) {
    for (int e = 0; e < etiCount; e++) {
      Bitmap overlap = etis[e] & *(dataSet->GetTransactionBitmap(r));
      int ones = overlap.CountOnes();
      if (ones >=  (int) ((1 - epsRow) * etis[e].CountOnes())) {
	dataSet->SetRowUnavailable(r);
	//cout << "row #" << r << " is covered" << endl;
	break;
      }
      //      cout << "ones: " << ones << endl;
    }
  }

  return anyAdded;
}

void ETIGreedy::PrintItems(Bitmap *items) 
{
  int itemCount = dataSet->GetItemCount();
    int tc = 0;

  for (int i = 0; i < itemCount; i++) {
    if (items->GetBit(i)) {
      tc++;
    }
  }

  if (tc == 1)
	  return;

  for (int i = 0; i < itemCount; i++) {
    if (items->GetBit(i)) {
      cout << dataSet->GetActualIndex(i) << " ";
    }
  }
  cout << endl;
}

bool ETIGreedy::HasStrongSupport(Bitmap items) 
{
	int support = 0;
	int rows = dataSet->GetTransactionCount();
	for (int r = 0; r < rows; r++) {
		Bitmap overlap = items & *(dataSet->GetTransactionBitmap(r));
		if (overlap.CountOnes() >= (int) ((1 - epsRow) * items.CountOnes())) {
			support++;
		}
	}

	return support >= pruningSupport[0];
}

bool ETIGreedy::AddETIToList(Bitmap items)
{
  // test to see if this item isn't already covered
  int size = etis.size();
  for (int i = 0; i < size; i++) {
    if (items.ContainedIn(&(etis[i]))) {      
      return false;      
    }
  }
    
  etis.push_back(items);
  
  return true;
}

double ETIGreedy::ComputeSigma(Bitmap *items, int *zeros, int lambda, bool debug)
{
  int transCount = dataSet->GetTransactionCount();
  int itemSetSize = items->CountOnes();
  int minimumSupport = (int) ceil(minSup / lambda);

  // counts the number of transactions containing any given number of ones
  // C_i in the ETI paper 
  ComputeZeros(items, zeros);
  int *oneCount = new int[itemSetSize + 1];
  memset(oneCount, 0, sizeof(int) * (itemSetSize + 1));
  for (int trans = 0; trans < transCount; trans++) {
    oneCount[itemSetSize - zeros[trans]]++;    
  }

  // find the value of t in the ETI paper (in Lemma 2)
  int *cumOneCount = new int[itemSetSize + 1];
  for (int i = 0; i <= itemSetSize; i++) {
    cumOneCount[i] = 0;
    for (int j = i; j <= itemSetSize; j++) {
      cumOneCount[i] += oneCount[j];
    }
  }
  int t = 0;
  while (cumOneCount[t] >= pruningSupport[0] && t < itemSetSize) {
    t++;
  }
  if (t > itemSetSize || cumOneCount[t] < pruningSupport[0]) {
    t--;
  }
  
  // find ct'
  int ctPrime = -minimumSupport;
  for (int i = t; i <= itemSetSize; i++) {
    ctPrime += oneCount[i];
  }  

  // find the sum in the equation given for sigma
  int sum = 0;
  for (int i = t + 1; i <= itemSetSize; i++) {
    sum += (itemSetSize - i) * oneCount[i];
  }

  if (debug) {
    cout << "------------------------" << endl;
    cout << "finding sigma for: " << *items;

    for (int i = 0; i <= itemSetSize; i++) {
      cout << oneCount[i] << " ";
    }
    cout << endl;
    cout << "t: " << t << ", oneCount[t]: " << oneCount[t] << endl;
  }

  double sigma = 
    (double) (sum + (itemSetSize - t) * (oneCount[t] - ctPrime)) /
    (minimumSupport * itemSetSize);
  
  delete [] oneCount;
  delete [] cumOneCount;

  return sigma;
}



double ETIGreedy::ComputeCoverage(Bitmap *items, int extend, int *zeros)
{
  int transCount = dataSet->GetTransactionCount();
  int itemSetSize = items->CountOnes();
  int zerosAllowed = (int) (itemSetSize * epsCol);
  int zeroCount = 0;
  int supportCount = 0; 

  // counts the number of transactions containing any given number of ones
  // C_i in the ETI paper 
  ComputeZeros(items, zeros);
  int *oneCount = new int[itemSetSize + 1];
  memset(oneCount, 0, sizeof(int) * (itemSetSize + 1));
  for (int trans = 0; trans < transCount; trans++) {
    oneCount[itemSetSize - zeros[trans]]++;
  }
  
  // find the value of t in the ETI paper (in Lemma 2)
  int t = 0;
  int cumCount = 0;
  while (transCount - cumCount >= pruningSupport[0]) {
    cumCount += oneCount[t];
    t++;
  }

  // find the values of Z_i from the ETI paper
  int *zeroCounters = new int[itemSetSize + 1];
  memset(zeroCounters, 0, sizeof(int) * (itemSetSize + 1));
  for (int trans = 0; trans < transCount; trans++) {
    if (!dataSet->ValueAt(trans, extend)) {
      int ones = itemSetSize - zeros[trans];
      zeroCounters[ones]++;
    }
  }
  
  // for each of the supporting transactions, find the number of zeros
  // in that transaction  
  for (int trans = 0; trans < transCount; trans++) {
    if (zeros[trans] <= zerosAllowed) {
      zeroCount += zeros[trans];
      supportCount++;
    }
  }

  // calculate the coverage
  int zeroSum = 0;
  for (int i = t + 1; i <= itemSetSize; i++) {
    zeroSum += zeroCounters[i];
  }
  int max = zeroCounters[t] - oneCount[t];
  max = (max > 0 ? max : 0);
  
  delete [] oneCount;
  delete [] zeroCounters;

  items->SetBit(extend, false);

  //  int blockSize = itemSetSize * supportCount;
  double numerator = zeroSum + max;
  double denominator = minSup;
  double coverage = 1 - (numerator / denominator);
  // cout << "coverage: " << coverage << endl;
  return coverage;
}

/* computes the pruning support threshold for each
 * itemset length (below which itemsets should be pruned)
 */
void ETIGreedy::ComputePruningSupport()
{
  int maxLength = dataSet->GetItemCount();
  int trans = dataSet->GetTransactionCount();
  pruningSupport = new int[maxLength + 1];
  pruningSupport[0] = minSup;
  //cerr << "support for " << "ETI-itemsets: " << pruningSupport[0] << endl;

  for (int length = 1; length <= maxLength; length++) {
    double factor =
      trans - (trans * length * epsCol) / (floor(length * epsCol) + 1);    
    int potentialSupport = 
     (int) ceil((double) minSup / trans * (factor > 0 ? factor : 0)); 
    pruningSupport[length] =       
      (potentialSupport ? potentialSupport : 1);
  }
  /*int maxLength = dataSet->GetItemCount();
  int trans = dataSet->GetTransactionCount();
  pruningSupport = new int[maxLength + 1];
  pruningSupport[0] = minSup;
  //cerr << "support for " << "ETI-itemsets: " << pruningSupport[0] << endl;

  for (int length = 1; length <= maxLength; length++) {
    double factor =
      trans - (trans * length * epsCol) / (floor(length * epsCol) + 1);    
    int potentialSupport = 
     (int) ceil((double) minSup / trans * (factor > 0 ? factor : 0)); 
    pruningSupport[length] = minSup;       
    //  (potentialSupport ? potentialSupport : 1);
  }*/
}

// for each of the transactions, counts how many zeros
// occur over the given set of indices (items)
void ETIGreedy::ComputeZeros(Bitmap *indices, int *zeros)
{
  int trans = dataSet->GetTransactionCount();
  memset(zeros, 0, sizeof(int) * trans);

  // for each item, see if it's in the set, if it is
  // then add that items zeros to the vector
  int itemSetSize = dataSet->GetItemCount();  
  for (int item = 0; item < itemSetSize; item++) {
    if (indices->GetBit(item)) {
      dataSet->AddItemToZeroCount(zeros, item);
    }
  }
}
